"""
Utilities for asynchronous LDAP operations.

HTTP endpoints typically trigger asynchronous LDAP requests
which return an ID for the operation being performed.
Results are then gathered in non-blocking mode.

Some shorthands are provided for common usages
like retrieving a unique result or waiting for an
operation to complete without results.
"""

from dataclasses import dataclass
from http import HTTPStatus
from typing import AsyncGenerator

from anyio import sleep
from fastapi import HTTPException
from ldap import (
    OPT_X_TLS_DEMAND,  # type: ignore
    OPT_X_TLS_NEVER,  # type: ignore
    OPT_X_TLS_NEWCTX,  # type: ignore
    OPT_X_TLS_REQUIRE_CERT,  # type: ignore
    SCOPE_BASE,  # type: ignore
    SCOPE_SUBTREE,  # type: ignore
    initialize,
)
from ldap.ldapobject import LDAPObject
from ldap.schema import SubSchema

from . import settings

# Constant to add technical attributes in LDAP search results
WITH_OPERATIONAL_ATTRS = ("*", "+")


@dataclass(frozen=True)
class LdapEntry:
    dn: str
    attrs: dict[str, list[bytes]]

    def attr(self, name: str) -> list[str]:
        return [v.decode() for v in self.attrs[name]]

    @property
    def hasSubordinates(self):
        return (
            self.attr("hasSubordinates") == ["TRUE"]
            if "hasSubordinates" in self.attrs
            else bool(self.attrs.get("numSubordinates"))
        )


sub_schema: SubSchema | None = None


def ldap_connect() -> LDAPObject:
    "Open an LDAP connection"

    url = settings.LDAP_URL
    connection = initialize(url)

    # #43 TLS, see https://stackoverflow.com/a/8795694
    if settings.USE_TLS or settings.INSECURE_TLS:
        cert_level = OPT_X_TLS_NEVER if settings.INSECURE_TLS else OPT_X_TLS_DEMAND

        connection.set_option(OPT_X_TLS_REQUIRE_CERT, cert_level)
        # See https://stackoverflow.com/a/38136255
        connection.set_option(OPT_X_TLS_NEWCTX, 0)
        if not url.startswith("ldaps://"):
            connection.start_tls_s()
    return connection


async def anonymous_user_search(connection: LDAPObject, username: str) -> str | None:
    try:
        return (
            await unique(
                connection,
                connection.search(
                    settings.BASE_DN,
                    SCOPE_SUBTREE,
                    settings.GET_BIND_DN_FILTER(username),
                ),
            )
        ).dn

    except HTTPException:
        pass  # No unique result


async def results(
    connection: LDAPObject, msgid: int
) -> AsyncGenerator[LdapEntry, None]:
    "Stream LDAP result entries without blocking other tasks"

    while True:
        r_type, r_data = connection.result(msgid=msgid, all=0, timeout=0)
        if r_type is None:  # Throttle to 100 results / second
            await sleep(0.01)
        elif r_data == []:  # Operation completed
            break
        else:
            yield LdapEntry(*r_data[0])  # type: ignore


async def unique(
    connection: LDAPObject,
    msgid: int,
) -> LdapEntry:
    "Asynchronously collect a unique result"

    res = None
    async for r in results(connection, msgid):
        if res is None:
            res = r
        else:
            connection.abandon(msgid)
            raise HTTPException(
                HTTPStatus.INTERNAL_SERVER_ERROR,
                "Non-unique result",
            )
    if res is None:
        raise HTTPException(HTTPStatus.NOT_FOUND, "Empty search result")
    return res


async def empty(
    connection: LDAPObject,
    msgid: int,
) -> None:
    "Asynchronously wait for an empty result"

    async for r in results(connection, msgid):
        connection.abandon(msgid)
        raise HTTPException(HTTPStatus.INTERNAL_SERVER_ERROR, "Unexpected result")


async def get_entry_by_dn(
    connection: LDAPObject,
    dn: str,
) -> LdapEntry:
    "Asynchronously retrieve an LDAP entry by its DN"

    return await unique(connection, connection.search(dn, SCOPE_BASE))


async def get_schema(connection: LDAPObject) -> SubSchema:
    global sub_schema
    # See: https://hub.packtpub.com/python-ldap-applications-part-4-ldap-schema/
    if sub_schema is None:
        result = await unique(
            connection,
            connection.search(
                settings.SCHEMA_DN,
                SCOPE_BASE,
                attrlist=WITH_OPERATIONAL_ATTRS,
            ),
        )
        sub_schema = SubSchema(result.attrs, check_uniqueness=2)
    return sub_schema



import logging

import click
import uvicorn
from uvicorn.config import LOG_LEVELS
from uvicorn.logging import ColourizedFormatter
from uvicorn.main import LEVEL_CHOICES

import ldap_ui

from . import settings


def print_version(ctx: click.Context, param: click.Parameter, value: bool) -> None:
    if value:
        click.echo(ldap_ui.__version__)
        ctx.exit()


@click.command()
@click.option(
    "-b",
    "--base-dn",
    type=str,
    default=settings.BASE_DN,
    help="LDAP base DN.  [default: Detect from root DSE]",
)
@click.option(
    "-h",
    "--host",
    type=str,
    default="127.0.0.1",
    help="Bind socket to this IP.",
    show_default=True,
)
@click.option(
    "-p",
    "--port",
    type=int,
    default=5000,
    help="Bind socket to this port (or 0 for any available port).",
    show_default=True,
)
@click.option(
    "-u",
    "--ldap-url",
    type=str,
    default=settings.LDAP_URL,
    help="LDAP directory connection URL.",
    show_default=True,
)
@click.option(
    "-l",
    "--log-level",
    type=LEVEL_CHOICES,
    default="info",
    help="Log level.",
    show_default=True,
)
@click.option(
    "--reload",
    is_flag=True,
    help="Watch for changes and reload?",
)
@click.option(
    "--version",
    is_flag=True,
    callback=print_version,
    expose_value=False,
    is_eager=True,
    help="Display the current version and exit.",
)
def main(base_dn, host, port, ldap_url, log_level, reload):
    logging.basicConfig(level=LOG_LEVELS[log_level])
    rootHandler = logging.getLogger().handlers[0]
    rootHandler.setFormatter(ColourizedFormatter(fmt="%(levelprefix)s %(message)s"))

    if base_dn is not None:
        settings.BASE_DN = base_dn

    if ldap_url is not None:
        settings.LDAP_URL = ldap_url

    uvicorn.run("ldap_ui.app:app", host=host, port=port, reload=reload)


if __name__ == "__main__":
    main()



import os
from typing import Optional

from starlette.config import Config

config = Config(".env")

# App settings
DEBUG = config("DEBUG", cast=lambda x: bool(x), default=False)
PREFERRED_URL_SCHEME = "https"
SECRET_KEY = os.urandom(16)

#
# LDAP settings
#
LDAP_URL = config("LDAP_URL", default="ldap:///")

# Directory base DN.
# If unset, auto-detection from the root DSE is attempted.
# This works under the following conditions:
# - The root DSE is readable with anonymous binding
# - `namingContexts` contains exactly one entry
# Otherwise, manual configuration is required.
BASE_DN = config("BASE_DN", default=None)

# DN to obtain the directory schema.
# If unset, auto-detection from the root DSE is attempted.
# This works if root DSE is readable with anonymous binding.
# Otherwise, manual configuration is required.
SCHEMA_DN = config("SCHEMA_DN", default=None)

USE_TLS = config(
    "USE_TLS", cast=lambda x: bool(x), default=LDAP_URL.startswith("ldaps://")
)

# DANGEROUS: Disable TLS host name verification.
INSECURE_TLS = config("INSECURE_TLS", cast=lambda x: bool(x), default=False)

#
# Binding
#


def GET_BIND_DN() -> Optional[str]:
    """
    Try to find a hard-wired DN from in the environment.
    If this is present and GET_BIND_PASSWORD returns something,
    the UI will NOT ask for a login.
    You need to secure it otherwise!
    """
    return config("BIND_DN", default=None)


def GET_BIND_PATTERN(username) -> Optional[str]:
    """
    Apply an optional user DN pattern for authentication
    from the environment,
    e.g. "uid=%s,ou=people,dc=example,dc=com".
    This can be used to authenticate with directories
    that do not allow anonymous users to search.
    """
    if config("BIND_PATTERN", default=None) and username:
        return config("BIND_PATTERN") % username


def GET_BIND_DN_FILTER(username) -> str:
    "Produce a LDAP search filter for the login DN"
    return SEARCH_PATTERNS[0] % username


def GET_BIND_PASSWORD() -> Optional[str]:
    "Try to determine the login password from the environment or request"
    pw = config("BIND_PASSWORD", default=None)
    if pw is not None:
        return pw

    pw_file = config("BIND_PASSWORD_FILE", default=None)
    if pw_file is not None:
        with open(pw_file) as file:
            return file.read().rstrip("\n")


#
# Search
#

# Attribute to search for user names
LOGIN_ATTR = config("LOGIN_ATTR", default="uid")

# Search users by a number of common attributes
SEARCH_PATTERNS = (
    "(%s=%%s)" % LOGIN_ATTR,
    "(cn=%s*)",
    "(gn=%s*)",
    "(sn=%s*)",
)

SEARCH_QUERY_MIN = config(
    "SEARCH_QUERY_MIN",  # Minimum length of query term
    cast=int,
    default=2,
)

SEARCH_MAX = config(
    "SEARCH_MAX",  # Maximum number of results
    cast=int,
    default=50,
)



"""
Simplistic ReST proxy for LDAP access.

Authentication is either hard-wired in the settings,
or else only HTTP basic auth is supported.

The backend is stateless, it re-connects to the directory on every request.
No sessions, no cookies, nothing else.
"""

import logging
from http import HTTPStatus

from fastapi import FastAPI, Request, Response
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.responses import JSONResponse
from fastapi.staticfiles import StaticFiles
from ldap import (  # type: ignore
    ALREADY_EXISTS,  # type: ignore
    INSUFFICIENT_ACCESS,  # type: ignore
    INVALID_CREDENTIALS,  # type: ignore
    NO_SUCH_OBJECT,  # type: ignore
    OBJECT_CLASS_VIOLATION,  # type: ignore
    UNWILLING_TO_PERFORM,  # type: ignore  # type: ignore
    LDAPError,  # type: ignore
)

from . import __version__, ldap_api, settings

# Main ASGI entry

app = FastAPI(debug=settings.DEBUG, title="LDAP UI", version=__version__)
app.include_router(ldap_api.api)
app.mount("/", StaticFiles(packages=["ldap_ui"], html=True))

app.add_middleware(GZipMiddleware, minimum_size=1000, compresslevel=5)


@app.middleware("http")
async def cache_buster(request: Request, call_next) -> Response:
    "Forbid caching of API responses"
    response = await call_next(request)
    if request.url.path.startswith("/api"):
        response.headers["Cache-Control"] = "no-cache, no-store, must-revalidate"
        response.headers["Pragma"] = "no-cache"
        response.headers["Expires"] = "0"
    return response


# API error handling

LDAP_ERROR_TO_STATUS = {
    ALREADY_EXISTS: HTTPStatus.CONFLICT,
    INSUFFICIENT_ACCESS: HTTPStatus.FORBIDDEN,
    INVALID_CREDENTIALS: HTTPStatus.UNAUTHORIZED,
    NO_SUCH_OBJECT: HTTPStatus.NOT_FOUND,
    OBJECT_CLASS_VIOLATION: HTTPStatus.BAD_REQUEST,
    UNWILLING_TO_PERFORM: HTTPStatus.FORBIDDEN,
}


@app.exception_handler(LDAPError)
def handle_ldap_error(request: Request, exc: LDAPError) -> Response:
    "General handler for LDAP errors"

    exc_type = type(exc)
    if exc_type is UNWILLING_TO_PERFORM:
        logging.critical("Need BIND_DN or BIND_PATTERN to authenticate")

    if exc_type is INVALID_CREDENTIALS:
        return Response(
            status_code=HTTPStatus.UNAUTHORIZED,
            headers={
                "WWW-Authenticate": 'Basic realm="Please log in", charset="UTF-8"'
            },
        )

    if exc_type not in LDAP_ERROR_TO_STATUS:
        # Unknown error --> log it since FastApi won't do it for us
        logging.exception("Error in %s %s:", request.method, request.url, exc_info=exc)

    cause = exc.args[0] if exc.args else {}
    desc = cause.get("desc", "LDAP error").capitalize()
    msg = f"{desc}" if "info" not in cause else f"{desc}: {cause['info'].capitalize()}"
    return JSONResponse(
        {"detail": [msg]},
        status_code=LDAP_ERROR_TO_STATUS.get(
            exc_type, HTTPStatus.INTERNAL_SERVER_ERROR
        ),
    )
